/* --- FlukaInit --- *
 * Initialize the fluka.@PARTICLES variable and the material. array
 */
FlukaInit: procedure expose fluka. material.
	particles = "HEAVYION -7 4-HELIUM -6 3-HELIUM -5 TRITON -4 DEUTERON -3 HEAVYION -2 OPTIPHOT -1"
	particles = particles "RAY 0 PROTON 1 APROTON 2 ELECTRON 3 POSITRON 4 NEUTRIE 5"
	particles = particles "ANEUTRIE 6 PHOTON 7 NEUTRON 8 ANEUTRON 9 MUON+ 10 MUON- 11"
	particles = particles "KAONLONG 12 PION+ 13 PION- 14 KAON+ 15 KAON- 16 LAMBDA 17"
	particles = particles "ALAMBDA 18 KAONSHRT 19 SIGMA- 20 SIGMA+ 21 SIGMAZER 22 PIZERO 23"
	particles = particles "KAONZERO 24 AKAONZER 25 NEUTRIM 27 ANEUTRIM 28 HEAVYION 30"
	particles = particles "ASIGMA- 31 ASIGMAZE 32 ASIGMA+ 33 XSIZERO 34 AXSIZERO 35"
	particles = particles "XSI- 36 AXSI+ 37 OMEGA- 38 AOMEGA+ 39 TAU+ 41 TAU- 42"
	particles = particles "NEUTRIT 43 ANEUTRIT 44 @LASTPAR 64 ALL-PART 201 ALL-CHAR 202 ALL-NEUT 203"
	particles = particles "ALL-NEGA 204 ALL-POSI 205 NUCLEONS 206 NUC&PI+- 207 ENERGY 208"
	particles = particles "PIONS+- 209 BEAMPART 210 EM-ENRGY 211 MUONS 212 E+&E- 213"
	particles = particles "AP&AN 214 KAONS 215 STRANGE 216 KAONS+- 217 HAD-CHAR 218"
	particles = particles "FISSIONS 219 HE-FISS 220 LE-FISS 221 NEU-BALA 222 HAD-NEUT 223"
	particles = particles "KAONS0 224 UNB-ENER 229 UNB-EMEN 230"

	fluka.@PARTICLES = particles

	/* ----------- Init Materials --------------- */
	/* predefined materials */
	premat =	"BLCKHOLE  -1            0" /*  1 */
	premat = premat "VACUUM    -1            0" /*  2 */
	premat = premat "HYDROGEN  0.0000837  1001" /*  3 */
	premat = premat "HELIUM    0.000166   2004" /*  4 */
	premat = premat "BERYLLIU  1.848      4009" /*  5 */
	premat = premat "CARBON    2.000      6012" /*  6 */
	premat = premat "NITROGEN  0.00117    7014" /*  7 */
	premat = premat "OXYGEN    0.00133    8016" /*  8 */
	premat = premat "MAGNESIU  1.740     12000" /*  9 */
	premat = premat "ALUMINUM  2.699     13027" /* 10 */
	premat = premat "IRON      7.874     26000" /* 11 */
	premat = premat "COPPER    8.960     29000" /* 12 */
	premat = premat "SILVER   10.500     47000" /* 13 */
	premat = premat "SILICON   2.329     14000" /* 14 */
	premat = premat "GOLD     19.320     79197" /* 15 */
	premat = premat "MERCURY  13.546     80000" /* 16 */
	premat = premat "LEAD     11.350     82000" /* 17 */
	premat = premat "TANTALUM 16.654     73181" /* 18 */
	premat = premat "SODIUM    0.971     11023" /* 19 */
	premat = premat "ARGON     0.00166   18000" /* 20 */
	premat = premat "CALCIUM   1.550     20000" /* 21 */
	premat = premat "TIN       7.310     50000" /* 22 */
	premat = premat "TUNGSTEN 19.300     74000" /* 23 */
	premat = premat "TITANIUM  4.540     22000" /* 24 */
	premat = premat "NICKEL    8.902     28000" /* 25 */

	do m=1 while premat <> ""
		parse var premat mat density zaid premat
		material.mat = m
		material.m = mat
		material.m.@density = density
		material.m.@zaid = zaid
		material.m.@used = 0
		material.m.@compound = zaid "1"
	end
	material.0 = m-1
	name = "@LASTMAT"
	material.m = name
	material.name = m-1
	drop premat

	fluka.@LINENO = 0
	fluka.@DEFINE.0 = 0
return

/* ------------- FlukaParticle ------------- */
FlukaParticle: procedure expose fluka.
	parse arg particle
	particle = strip(particle)
	if datatype(particle,"N") then return particle
	p = wordpos(particle,fluka.@PARTICLES)
	if p>0 then return word(fluka.@PARTICLES,p+1)
return particle

/* ------------ FlukaParticleName -------------- */
FlukaParticleName: procedure expose fluka.
	particle = arg(1)
	if ^datatype(particle,"N") then return particle
	p = wordpos(trunc(particle),fluka.@PARTICLES)
	if p>0 then return word(fluka.@PARTICLES,p-1)
return particle

/* ------------ _FlukaCheckName ---------- */
_FlukaCheckName:
	if ^datatype(name,"S") then do
		call lineout "<STDERR>",fluka.@LINENO":" arg(1)
		call lineout "<STDERR>","WARNING: Invalid" command
		return 1
	end
return 0

/* ------------- FlukaDefine --------------- */
FlukaDefine: procedure expose fluka.
	parse arg "#" command name val
	nest = fluka.@DEFINE.0
	if nest = 0 then
		action=1
	else
		action = fluka.@DEFINE.nest
	select
		when action=1 & command="define" then do
			if _FlukaCheckName(arg(1)) then return
			if val="" then val=1
			fluka.@DEFINE.name = val
		end
		when action=1 & command="undef" then do
			if _FlukaCheckName(arg(1)) then return
			fluka.@DEFINE.name=""
			/*drop fluka.@DEFINE.name*/
		end
		when command="if" then do
			if _FlukaCheckName(arg(1)) then return
			nest = nest + 1
			fluka.@DEFINE.0 = nest
			if action=1 then
				fluka.@DEFINE.nest = (fluka.@DEFINE.name<>"")
			else
				fluka.@DEFINE.nest = -1
			if value("verbose",,0)=1 then say fluka.@LINENO": #if" name "=" fluka.@DEFINE.nest
		end
		when command="elif" then do
			if _FlukaCheckName(arg(1)) then return
			if nest<=0 then do
				call lineout "<STDERR>",fluka.@LINENO":" arg(1)
				call lineout "<STDERR>","WARNING: Unexpected #elif"
				return
			end; else
			if fluka.@DEFINE.name<>"" & action=0 then
				fluka.@DEFINE.nest = 1
			else
				fluka.@DEFINE.nest = -1
			if value("verbose",,0)=1 then say fluka.@LINENO": #elif" name "=" fluka.@DEFINE.nest
		end
		when command="else" then do
			if nest<=0 then do
				call lineout "<STDERR>",fluka.@LINENO":" arg(1)
				call lineout "<STDERR>","WARNING: Unexpected #else"
				return
			end; else
			if action=1 then fluka.@DEFINE.nest = -1
			else
			if action=0 then fluka.@DEFINE.nest = 1
			if value("verbose",,0)=1 then say fluka.@LINENO": #else =" fluka.@DEFINE.nest
		end
		when command="endif" then do
			if value("verbose",,0)=1 then say fluka.@LINENO": #endif"
			if nest<=0 then do
				call lineout "<STDERR>",fluka.@LINENO":" arg(1)
				call lineout "<STDERR>","WARNING: Unexpected #endif"
				return
			end; else
				fluka.@DEFINE.0 = nest-1
		end
	end
return

/* ------------- FlukaReadLine ------------- */
FlukaReadLine: procedure expose fluka.
	trace o
	parse arg fin
	fluka.@COMMENT = ""
	do forever
		line = read(fin)
		if eof(fin) then return ""

		fluka.@LINENO = fluka.@LINENO + 1
		fluka.@LINE   = line

		/* Skip empty lines */
		if line="" then do
			call lineout "<STDERR>",fluka.@LINENO":" line
			call lineout "<STDERR>","WARNING: Empty line found!"
			iterate
		end

		/* Chop trailing CR if any */
		if right(line,1) == "0D"x then line=left(line,length(line)-1)

		/* Strip trailing blanks */
		line    = strip(line,"T")
		firstch = left(line,1)

		if firstch=="#" then do
			if value("preprocessor",,0) then
				call FlukaDefine line
			fluka.@COMMENT = fluka.@COMMENT"0A"x||"*-->" line
		end; else do
			if fluka.@DEFINE.0 > 0 then do
				nest = fluka.@DEFINE.0
				if fluka.@DEFINE.nest<>1 then do
					fluka.@COMMENT = fluka.@COMMENT"0A"x||"*-->" line
					iterate
				end
			end
			if firstch=="*" then
				fluka.@COMMENT = fluka.@COMMENT"0A"x||line
			else do
				pos=verify(line,xrange(' ','~'))
				if pos<>0 then do
					ch=substr(line,pos,1)
					call lineout "<STDERR>",fluka.@LINENO":" line
					call lineout "<STDERR>","Error: Invalid character",
						"found at col="pos "char="ch c2d(ch)
					exit 1
				end
				return strip(line,"T")
			end
		end
	end

/* ------------- FlukaFindCard ------------- */
FlukaFindCard: procedure expose fluka.
	parse arg fin,card
	do forever
		line = FlukaReadLine(fin)
                if eof(fin) then return ""
		if card = left(line,8) then return line
	end

/* --------- FlukaFreeFormat ----------- */
FlukaFreeFormat: procedure
	trace o
	line = arg(1)
	parse var line line "!" .
	sep=',;:/\'
	do i=2 to length(sep)
		s=substr(sep,i,1)
		line = changestr(s,line,',')
	end
	lineout = line
	do until lineout==line
		line = lineout
		lineout = changestr(" ,",line,",")
		lineout = changestr(", ",lineout,",")
	end
return changestr(' ',space(lineout),',')

/* --------- FlukaParseLine ----------- */
FlukaParseLine: procedure expose card extra what. sdum
	parse arg line,format
	if format=1 | left(format,1)="F" then do
		line = FlukaFreeFormat(line)
		parse var line card "," what.1 "," what.2 "," what.3 ",",
					what.4 "," what.5 "," what.6 "," sdum
		what.0 = 6
	end; else do
		parse var line card +8 extra +2,
			what.1 +10 what.2 +10 what.3 +10,
			what.4 +10 what.5 +10 what.6 +10 sdum
		if length(extra)<>2 then extra = '  '
		card = strip(card)
		do i=1 to 6
			if length(what.i) <> 10 then
				what.i = right(strip(what.i),10)
		end
		what.0 = 6
		sdum = strip(sdum)
	end
	do i=1 to what.0
		s = strip(what.i)
		if s="" then iterate
		if pos(left(s,1),"+-.0123456789")>0 &,
		   pos(substr(s,2,1),".0123456789dDeE")>0 then do
			what.i = translate(what.i,"EE","dD")
		end
	end
return

/* --------- FlukaParsePrimitive ----------- *
 * Arguments:	arg(1):	Line to parse
 *		arg(2): Format: 0=normal fixed format
 *				1=double precission fixed format
 *                              2=free format
 *		arg(3): Number of whats already read
 */
FlukaParsePrimitive: procedure expose primitive name what. sdum
	trace o
	parse arg line, format, nwhat
	startwhat = nwhat
	select
		when format=0 | format="" then do
			parse var line 3 primitive +3 name +5 line
			if nwhat>0 & (primitive<>"" | name<>"") then do
				call lineout "<STDERR>","ERROR in line:'"primitive||name||line"'"
				call lineout "<STDERR>","      found body/name where it was expecting blank"
				exit 0
			end
			do i=nwhat+1 to nwhat+6
				parse var line what.i +10 line
			end
			sdum = line
			nwhat = nwhat+6
			len=10
		end
		when format=1 then do
			parse var line 3 primitive +3 name +5 line
			if nwhat>0 & (primitive<>"" | name<>"") then do
				call lineout "<STDERR>","ERROR in line:'"primitive||name||line"'"
				call lineout "<STDERR>","      found body/name where it was expecting blank"
				exit 0
			end
			do i=nwhat+1 to nwhat+3
				parse var line what.i +22 line
			end
			sdum  = line
			nwhat = nwhat+3
			len = 22
		end
		when format=2 then do
			len = 0
			line = FlukaFreeFormat(line)
			if nwhat=0 then parse var line primitive "," name "," line
			sdum = ""
			do while line<>""
				nwhat = nwhat+1
				parse var line what.nwhat "," line
			end
			/* Maybe it needs to be expanded in more lines */
		end
		otherwise
			call lineout "<STDERR>","Error: In ParseFlukaPrimitiveLine Unknown geometry format" format
			exit -1
	end
	do i=startwhat+1 to nwhat
		s = strip(what.i)
		if s="" then iterate
		if pos(left(s,1),"+-.0123456789")>0 &,
		   pos(substr(s,2,1),".0123456789dDeE")>0 then do
			what.i = translate(what.i,"EE","dD")
		end
		if len>0 & length(what.i)<>len then
			what.i = right(strip(what.i),len)
		if datatype(what.i,"N") then
			what.i = what.i + 0.0
	end
	sdum = strip(sdum)
return nwhat

/* --------- FlukaReadPrimitive ----------- */
FlukaReadPrimitive: procedure expose primitive name what. sdum fluka.
	trace o
	parse arg fgeo, geoFormat, cont

	if cont then comment = fluka.@COMMENT
		else what.0 = 0

	line = FlukaReadLine(fgeo)
	what.0 = FlukaParsePrimitive(line,geoFormat,what.0)

	if cont then do
		if fluka.@COMMENT<>"" then do
			if comment<>""  then
				comment=comment"0A"x||fluka.@COMMENT
			else
				comment=fluka.@COMMENT
		end
		fluka.@COMMENT = comment
	end
return line

/* --------- FlukaParseRegion ----------- */
FlukaParseRegion: procedure expose region region_expr region_n
	trace o
	parse arg line, format
	if format=2 then do
		parse var line region region_n region_expr
		region_expr = changestr("(|",space(region_expr,0),"(")
		if left(region_expr,1)=="|" then region_expr=substr(region_expr,2)

		if value("expandGeom",,0) then
			region_expr = CSGNormalize(region_expr)
	end; else do
		parse var line 2 region +5 region_n +4 line
		region_expr = ""
		do while line <> ""
			parse var line or +2 vol +5 line
			if or="OR" then
				region_expr = region_expr "|"
			else
			if or^=="  " then do
				call lineout "<STDERR>","ERROR: Region defintion error in OR position found='"or"'"
			end
			vol = strip(vol)
			prefix = left(vol,1)
			if prefix<>"-" & prefix<>"+" then vol = '+'vol
			region_expr = region_expr vol
		end
	end
	region = strip(region)
	region_expr = space(region_expr)
return region_expr

/* --------- FlukaReadRegion ----------- */
FlukaReadRegion: procedure expose region region_expr region_n fluka.
	parse arg fgeo, format

	line = FlukaReadLine(fgeo)
	comment = fluka.@COMMENT
	orgline = line
	if line<>"END" then do
		do forever
			fpos   = seek(fgeo)
			lineno = fluka.@LINENO
			line2  = FlukaReadLine(fgeo)
			if format=2 then do
				line2 = strip(line2)
				if pos(left(line2,1),"+-|(")>0 then do
					line = line || line2
					orgline = orgline"0A"x||line2
				end; else
					leave
			end; else do
				if left(line2,5)="" then do
					/* number of operation */
					len = length(line)
					if (len-10)//7 <> 0 then do
						n = (len-10)%7+1
						line = left(line,n*7+10)
					end
					line = line||substr(line2,11)
					orgline = orgline"0A"x||line2
				end; else
					leave
			end
			if fluka.@COMMENT<>"" then do
				if comment<>""  then
					comment=comment"0A"x||fluka.@COMMENT
				else
					comment=fluka.@COMMENT
			end
		end
		fluka.@LINENO = lineno
		fluka.@LINE   = line
		call seek fgeo,fpos
	end

	fluka.@COMMENT = comment

	call FlukaParseRegion line,format
return orgline

/* --------- FlukaCreateLine ----------- */
FlukaCreateLine: procedure expose card extra what. sdum
	parse arg len
	if len="" then len=10

	if length(card)<>8 then card=left(card,8)
	if length(extra)<>2 then extra = '  '

	multiline = ""
	line = card||extra

	do i=1 to what.0
		if datatype(what.i,"N") then
			what.i = FlukaFormat(what.i,len)
		line = line||right(strip(what.i),len)

		if length(line)>=70 then do
			if sdum<>"" then
				multiline = multiline||line||strip(sdum)
			else
				multiline = multiline||strip(line)
			if i<what.0 then do
				multiline=multiline"0A"x
				sdum = " &"
				line = card||extra
			end; else
				line = ""
		end
	end
	line = strip(line)
	if multiline<>"" then
		return multiline||line
	else
		return line

/* --------- FlukaCreatePrimitiveLine ----------- */
FlukaCreatePrimitiveLine: procedure expose primitive name what. sdum
	parse arg format

	multiline = ""
	if format<=1 then do
		line = "  "||left(primitive,3)||right(name,5)
		if format=0 then do
			len = 10
			maxlen = 70
		end; else do
			len = 22
			maxlen = 76
		end
	end; else do
		maxlen = 72
		line = primitive left(name,10)
	end

	do i=1 to what.0
		what = FlukaFormat(what.i)
		if format<=1 then
			what = right(what,len)
		else
			what = " "what

		if length(line)+length(what)>maxlen then do
			if multiline<>"" then
				multiline = multiline"0A"x||line
			else
				multiline = line
			line = copies(" ",10)
		end
		line = line || what
	end
	if multiline<>"" then
		return multiline"0A"x||line
	else
		return line
return

/* --- Format a Fluka number --- */
FlukaFormat: procedure
/*	trace o*/

	/* In the simplest case the first is the only argument. */
	Number = strip(arg(1))
	if arg(2,"E") then
		Len = arg(2)
	else
		Len = digits()
	Exp    = arg(3,"E")

	if ^datatype(Number,"N") then return Number
	if length(Number)<Len then do
		Number = translate(Number,"DDD","eEd")
		hasExp = pos("D",Number)<>0
		if Exp then do
			if hasExp then return Number
		end; else
		if pos(".",Number)<>0 | hasExp then return Number
	end
	if Number=="0" then
		if Exp	then return "0.D0"
			else return "0.0"

	if Len<5 | Len>22 then do
		call lineout "<STDERR>","ERROR: FlukaFormat invalid length"
		return Number
	end

	/* Dissect the Number. It is in the normal Rexx format. */
	parse upper var Number Mantissa 'E' Exponent
	if Exponent == '' then Exponent = 0

	Sign = 0
	if left(Mantissa,1) == '-' then do
		Sign = 1
		Mantissa = substr(Mantissa,2)
	end; else
	if left(Mantissa,1) == '+' then
		Mantissa = substr(Mantissa,2)

	parse var Mantissa Befo '.' Afte

	/* Count from the left for the decimal point. */
	Point = length(Befo)

	/* Make this a Number without a point. */
	Integer = Befo||Afte

	/* Remove leading and trailing zeros */
	p = verify(Integer,"0")
	if p=0 then do
		if Exp	then return "0.D0"
			else return "0.0"
	end
	Integer = substr(Integer,p)
	Point = Point-p+1

	p = verify(reverse(Integer),"0")
	if p>1 then Integer = left(Integer,length(Integer)-p+1)

	Exponent = Exponent + Point

	/* Correct the numeric digits */
	oldnd = digits()
	numeric digits 22

	/* Cannot handle more than 15 digits */
	if length(Integer)>15 then do
		r = substr(Integer,16,1)
		Integer=left(Integer,15)
		if r>='5' then Integer = Integer+1
	end

	/* Now the number is described by
	 *	Sign 0.Integer "E" Exponent
	 */

	/* Make space for sign */
	if sign then Len = Len-1

	do forever
		/* Minimum length representation of a number
		 * Length = Length of Integer + 1 for Dot if needed (no exponent)
		 *	    + (2-4) for exponent
		 *	Exponent can be in the following forms
		 *		nothing if dot can placed inside Integer
		 *		E#	2
		 *		E##	3
		 *		E-#	3
		 *		E-##	4
		 *	Integer is given as  0.Integer
		 */
		if Exp then
			mNum = insert(".",Integer,1)||"D"||Exponent-1
		else select
			when Exponent=-2 then mNum = ".00"Integer
			when Exponent=-1 then mNum = ".0"Integer
			when Exponent=0  then mNum = "."Integer
			when Exponent=1  then mNum = insert(".",Integer,1)
			when Exponent>1 & Exponent<=Length(Integer) then
					mNum = insert(".",Integer,Exponent)
			when Exponent>1 & Exponent<=Length(Integer)+2 then
					mNum = left(Integer,Exponent,"0")"."
			when Exponent>1 & Exponent>Length(Integer) & Exponent+1<Len then
					mNum = Integer||copies("0",Exponent-Length(Integer))||"."
			otherwise
				mNum = insert(".",Integer,1)||"D"||Exponent-1
		end
		/*
		say ">F>" mNum
		say "LENGTH=" length(mNum) Len
		say "DIFF=" length(mNum)-Len
		*/

		Diff = length(mNum)-Len
		if Diff<=0 then leave
		if Diff<=2 then do
			r = right(Integer,1)
			Integer = left(Integer,length(Integer)-1)
		end; else do
			r = substr(Integer,length(Integer)-Diff+1,1)
			Integer = left(Integer,length(Integer)-Diff)
			/*say "R="R "I="Integer*/
		end
		if r>='5' then Integer = Integer+1

		/* Cut trailing zeros */
		p = verify(reverse(Integer),"0")
		if p=0 then do
			numeric digits oldnd
			if Exp	then return "0.D0"
				else return "0.0"
		end
		Integer = left(Integer,length(Integer)-p+1)
	end
	numeric digits oldnd
	if Sign then mNum = "-"mNum
return mNum
